# System
import json
from os import path, pardir, utime
from urlparse import urlunparse
from collections import namedtuple


class JsonFileDict:
    """
    Maintain a dictionary as a JSON object in a file
    with utility methods for manipulating the model
    """

    def __init__(self, filepath):
        self.filepath = filepath

        # Check for valid JSON, or create a new JSON object
        try:
            with open(self.filepath, 'r') as json_file:
                json.load(json_file)
        except (ValueError, IOError):
            with open(self.filepath, 'w') as json_file:
                json.dump({}, json_file)

    def all(self):
        """
        Get the whole JSON object from the file
        """

        with open(self.filepath, 'r') as json_file:
            return json.load(json_file)

    def overwrite(self, new_dictionary):
        """
        Overwrite the whole JSON object with a new one
        """

        with open(self.filepath, 'w') as json_file:
            json.dump(new_dictionary, json_file)

    def update(self, incoming_variables):
        """
        Add or update properties on the JSON object
        """

        variables = self.all()

        variables.update(incoming_variables)

        self.overwrite(variables)

    def get(self, variable_name):
        """
        Retrieve dictionary item from JSON dict
        """

        return self.all()[variable_name]

    def delete(self, variable_name):
        """
        Delete an item from the dictionary
        """

        json_dict = self.all()
        deleted = False

        if variable_name in json_dict:
            del json_dict[variable_name]
            self.overwrite(json_dict)
            deleted = True

        return deleted

    def set(self, variable_name, variable_value):
        """
        Create or update a value in the JSON dict
        in the file
        """

        json_dict = self.all()
        json_dict[variable_name] = variable_value
        self.overwrite(json_dict)


def literal(**kw):
    """
    Create an "object literal"
    by creating a named tuple with keyword arguments
    """

    return namedtuple('literal', kw)(**kw)


def parent_dir(dir_path):
    """
    Get the parent dir from of the top directory for a path,
    regardless of whether the path ends in a filename or not

    > parent_dir('/my/favourite/directory')
    '/my/favourite'

    > parent_dir('/my/favourite/directory/my-file.txt')
    '/my/favourite'
    """

    if not path.isdir(dir_path):
        dir_path = path.dirname(dir_path)

    return path.abspath(path.join(dir_path, pardir))


def build_url(
    domain,
    scheme='http', port=None, username=None, password=None,
    path='', params='', query='', fragment=''
):
    """
    Build a URL from its parts

    This is just a proxy for urlunparse,
    but with more descriptive parameter names

    > build_url(domain='example.com', port=8080, path='users')
    'http://example.com:8080/users'
    """

    return urlunparse(
        (
            scheme,
            build_url_host(domain, port, username, password),
            path,
            params,
            query,
            fragment
        )
    )


def build_url_host(domain, port=None, username=None, password=None):
    """
    Build the host part of a URL from a domain, port, username and password

    > build_url_host('example.com')
    'example.com'

    > build_url_host('example.com', 8080)
    'example.com:8080'

    > build_url_host('example.com', 8080, 'robin')
    'robin@example.com:8080'

    > build_url_host('example.com', 8080, 'robin', 'mypassword')
    'robin:mypassword@example.com:8080'
    """

    host = domain

    if port:
        host = "{0}:{1}".format(domain, port)

    if username:
        credentials = username

        if password:
            credentials = "{0}:{1}".format(
                username, password
            )

        host = "{0}@{1}".format(credentials, host)

    return host


def add_ansible_config(charm_dir, config_data):
    """
    Collect config settings from an ansible role
    by reading both 'defaults/main.yml'
    and 'vars/main.yml'
    and appending their data to config_data
    """

    # Local imports - 'cos they weren't ready earlier
    from jinja2 import Environment, FileSystemLoader

    # Setup template parser environment
    wsgi_role_path = path.join(charm_dir, 'roles', 'wsgi-app')
    template_env = Environment(loader=FileSystemLoader(wsgi_role_path))

    config_data = update_from_yaml_template(
        'defaults/main.yml', template_env, config_data
    )

    config_data = update_from_yaml_template(
        'vars/main.yml', template_env, config_data
    )

    return config_data


def update_from_yaml_template(template_path, parser_env, data):
    """
    Given a path to a yaml file and a template parser environment
    add all data in the yaml file to the provided data dictionary
    """

    import yaml
    from jinja2 import meta, UndefinedError

    # Find variables in template
    content = parser_env.loader.get_source(parser_env, template_path)
    parsed_template = parser_env.parse(content[0])
    template_variables = meta.find_undeclared_variables(parsed_template)

    # Find undeclared variables not already in context
    undefined_variables = list(set(template_variables) - set(data))

    # Parse template
    template = parser_env.get_template(template_path)
    yaml_content = template.render(data)
    new_variables = yaml.load(yaml_content)
    data.update(new_variables)

    # If any new variables were in the undeclared variables
    list_intersection = filter(
        lambda x: x in undefined_variables, new_variables
    )

    if list_intersection:
        data = update_from_yaml_template(template_path, parser_env, data)
    elif undefined_variables:
        raise UndefinedError('Undefined variables: ' + str(undefined_variables))

    return data


def items_are_not_empty(test_dictionary, items_to_test):
    """
    For a list of keys, check all items in a dictionary with those key names
    are not empty

    > items_are_not_empty({'a': 'hi'}, ['a'])
    True

    > items_are_not_empty({'a': 'hi', 'b': ''}, ['a', 'b'])
    False

    > items_are_not_empty({'a': 'hi', 'b': ''}, ['a', 'c'])
    False

    > items_are_not_empty({'a': 'hi', 'b': '', 'c': 1}, ['a', 'c'])
    True

    """

    # A list of booleans representing whether items evaluate to True
    item_booleans = [bool(test_dictionary.get(k)) for k in items_to_test]

    # Return True only if all items are present and True
    return reduce(lambda x, y: x and y, item_booleans)


def dequote(s):
    """
    If a string has single or double quotes around it, remove them.
    If a matching pair of quotes is not found, return the string unchanged.
    """

    if (
        s.startswith(("'", '"')) and s.endswith(("'", '"'))
        and (s[0] == s[-1])  # make sure the pair of quotes match
    ):
        s = s[1:-1]
    return s

def touch(filename, times=None):
    """
    Update the time on a file,
    creating it if it doesn't exist
    (From: http://stackoverflow.com/a/1160227)
    """

    with open(filename, 'a'):
        utime(filename, times)
