#!/usr/bin/env python

# System
import os
import re
import sys
from base64 import b64encode
from datetime import datetime

# Local
import sh
from helpers import (
    add_ansible_config,
    build_url,
    dequote,
    items_are_not_empty,
    parent_dir,
    JsonFileDict
)
import charmhelpers.contrib.ansible
from charmhelpers.core.hookenv import (
    close_port,
    config,
    Hooks,
    local_unit,
    open_port,
    relation_get,
    relation_ids,
    relation_set,
    relations,
    UnregisteredHookError
)
from charmhelpers.core.host import log
from charmhelpers.payload.execd import execd_preinstall

# Globals (unfortunately)
charm_dir = parent_dir(__file__)
unit_dir = parent_dir(charm_dir)
state_dir = os.path.join(unit_dir, 'state')
env_file = os.path.join(state_dir, 'environment-variables.json')
environment_variables = JsonFileDict(env_file)

# Hooks helper for the direct python hooks
# (ansible hooks are run by ansible_hooks)
hooks = Hooks()

# Create the hooks helper which automatically registers the
# required hooks based on the available tags in your playbook.
# By default, running a hook (such as 'config-changed') will
# result in running all tasks tagged with that hook name.
ansible_hooks = charmhelpers.contrib.ansible.AnsibleHooks(
    playbook_path='playbook.yml'
)


@ansible_hooks.hook('upgrade-charm')
def upgrade():
    """
    Before running normal install code,
    copy any environment variables out of charm_cache/env.json
    into the new location ../state/environment-variables.json.
    Keeping them outside the charm means they can persist
    through charm upgrades.
    """

    old_env_file = os.path.join(charm_dir, 'charm_cache', 'env.json')

    if os.path.isfile(old_env_file):
        old_environment_variables = JsonFileDict(old_env_file)
        existing_env = old_environment_variables.all()
        existing_env.update(environment_variables.all())
        environment_variables.update(existing_env)

    ansible_install()


@ansible_hooks.hook('install')
def ansible_install():
    """
    Before anything else
    - do preinstall hooks
    - setup ansible
    """

    execd_preinstall()
    charmhelpers.contrib.ansible.install_ansible_support(from_ppa=True)


@hooks.hook('wsgi-file-relation-changed')
def wsgi_relation():
    """
    Setup relation for serving the WSGI file (e.g. gunicorn)

    Sets a whole bunch of relation settings
    including log file locations and environent variables
    """

    log('Hook function: wsgi_relation')

    config_data = ansible_config()

    access_log_path = os.path.join(
        config_data['log_dir'],
        config_data['app_label'] + '-access.log'
    )
    error_log_path = os.path.join(
        config_data['log_dir'],
        config_data['app_label'] + '-error.log'
    )

    env_dictionary = environment_variables.all()
    env_list = ["{0}={1}".format(k, v) for k, v in env_dictionary.items()]
    env_string = " ".join(env_list)

    wsgi_relation_settings = {
        'project_name': config_data.get('app_label', ''),
        'working_dir': os.path.join(
            config_data.get('code_dir', ''),
            'current'
        ),
        'python_path': config_data.get('python_path', ''),
        'wsgi_user': config_data.get('wsgi_user', ''),
        'wsgi_group': config_data.get('wsgi_group', ''),
        'port': config_data.get('listen_port', ''),
        'wsgi_access_logfile': access_log_path,
        'wsgi_wsgi_file': config_data.get('wsgi_application', ''),
        'wsgi_extra': '--error-logfile=' + error_log_path,
        'env_extra': env_string,
        'timestamp': datetime.now().isoformat()
    }

    # Set these settings on any wsgi-file relations
    for relation_id in relation_ids('wsgi-file'):
        log(
            'Setting wsgi-file relation settings: {}'.format(
                wsgi_relation_settings
            )
        )

        relation_set(
            relation_id=relation_id,
            **wsgi_relation_settings
        )

        # Relation changed - re-run update target
        update_target()

    open_port(config_data['listen_port'])


@hooks.hook('wsgi-relation-broken')
def wsgi_relation_broken():
    """
    When WSGI relation (e.g.: gunicorn) goes away
    """

    log('Hook function: wsgi_relation_broken')

    config_data = ansible_config()

    close_port(config_data['listen_port'])


@hooks.hook('pgsql-relation-changed')
def pgsql_relation():
    """
    Setup relation to a postgresql database
    (will replace any other database relations)
    """

    log('Hook function: pgsql_relation')

    host = relation_get("host")

    if 'pgsql' in relations() and host:
        link_database(
            scheme='postgresql',
            database_host=host,
            port=relation_get("port"),
            username=relation_get("user"),
            password=relation_get("password"),
            database_name=relation_get("database")
        )


@hooks.hook('mongodb-relation-changed')
def mongodb_relation():
    """
    Setup relation to a mongodb database
    (will replace any other database relations)
    """

    log('Hook function: mongodb_relation')

    host = relation_get("hostname")

    if 'mongodb' in relations() and host:
        link_database(
            scheme='mongodb',
            database_host=host,
            port=relation_get("port"),
            variable_name='MONGO_URL'
        )


@hooks.hook('pgsql-relation-broken')
def unlink_pgsql():
    unlink_database('DATABASE_URL')


@hooks.hook('mongodb-relation-broken')
def unlink_mongo():
    unlink_database('MONGO_URL')


def unlink_database(variable_name):
    """
    Remove "DATABASE_URL" environment variable
    """

    log('Function: unlink_database')

    if environment_variables.delete('DATABASE_URL'):
        # Reset wsgi relation settings
        wsgi_relation()


@hooks.hook('webservice-relation-changed')
def webservice_relation():
    """
    Create "WEBSERVICE_URL" environment variable from relation
    """

    log('Function: webservice_relation')
    http_protocol = config('webservice_protocol')
    port = config('webservice_port') or relation_get("port")
    address = relation_get('servername') or relation_get('private-address')
    hostname = relation_get('hostname') or address

    # If hostname is IP address or FQDN, use it
    # otherwise use private_address
    ip_regex = re.compile(
        (
            r"^(([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\.)"
            r"{3}([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])$"
        )
    )
    hostname_regex = re.compile(
        (
            r"^(([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]*[a-zA-Z0-9])\.)+"
            r"([A-Za-z0-9]|[A-Za-z0-9][A-Za-z0-9\-]*[A-Za-z0-9])$"
        )
    )

    if hostname_regex.match(hostname) or ip_regex.match(hostname):
        domain = hostname
    else:
        domain = address

    webservice_url = build_url(
        scheme=http_protocol,
        domain=domain,
        port=port
    )

    environment_variables.set('WEBSERVICE_URL', webservice_url)

    # Relation changed - re-run update target
    update_target()

    # Reset wsgi relation settings
    wsgi_relation()


@hooks.hook('webservice-relation-broken')
def unlink_webservice():
    """
    Remove "WEBSERVICE_URL" environment variable
    """

    log('Function: unlink_database')

    if environment_variables.delete('WEBSERVICE_URL'):
        # Reset wsgi relation settings
        wsgi_relation()


@hooks.hook('log-archive-relation-changed')
def log_archive_relation_changed():
    """
    When WSGI relation (e.g.: gunicorn) goes away
    """

    log('Hook function: log_archive_relation_changed')

    relation_set(logdirs=b64encode(ansible_config()['log_dir']))
    relation_set(archive_after='1')


def update_env():
    'Save any environment variables as JSON'

    env_vars_string = config('environment_variables')

    if env_vars_string:
        for env_var_string in env_vars_string.split(' '):
            key, value = env_var_string.split('=')
            value = dequote(value)
            environment_variables.set(key, value)


@hooks.hook('install', 'upgrade-charm')
def install():
    """
    - Install ansible
    - Create the cache directory

    The hook() helper decorating this install function ensures that after this
    function finishes, any tasks in the playbook tagged with install are
    executed.
    """

    log('Hook function: install')

    update_env()


@hooks.hook('config-changed')
def config_changed():
    """
    Run everything which should be updated when config changes
    """
    update_env()
    wsgi_relation()
    update_target()
    pgsql_relation()
    mongodb_relation()


# Helper functions
# ===

def ansible_config():
    """
    Build ansible config data, extending current config_data
    and with the local unit name
    """

    config_data = config()
    config_data['local_unit'] = local_unit()
    return add_ansible_config(charm_dir, config_data)


def update_target():
    """
    Run the "update-charm" make target within the project
    """

    log('Hook function: update_target')

    config_data = ansible_config()

    required_configs = [
        'build_label',
        'archive_filename',
        'current_code_dir',
        'update_make_target'
    ]

    # Check all required configs are set
    if (
        items_are_not_empty(config_data, required_configs) and
        os.path.isdir(config_data['current_code_dir'])
    ):
        # Ensure make is installed
        apt_output = sh.apt_get.install('make')
        log('Installed make:')
        log(str(apt_output))

        # Execute make target with all environment variables
        make_output = sh.make(
            config_data['update_make_target'],
            directory=os.path.join(config_data['current_code_dir']),
            _env=environment_variables.all()
        )

        log('Make output:')
        log(str(make_output))


def link_database(
    scheme,
    database_host,
    port='',
    username='',
    password='',
    database_name='',
    variable_name='DATABASE_URL'
):
    """
    Create a link with a database by setting an environment variable
    The variable will be "DATABASE_URL" by default
    """

    database_url = build_url(
        scheme=scheme,
        domain=database_host,
        port=port,
        username=username,
        password=password,
        path=database_name
    )

    environment_variables.set(variable_name, database_url)

    # Relation changed - re-run update target
    update_target()

    # Reset wsgi relation settings
    wsgi_relation()


if __name__ == "__main__":
    ansible_failed = False

    # Run ansible hooks first
    try:
        ansible_hooks.execute(sys.argv)
    except UnregisteredHookError:
        ansible_failed = True

    # Then try regular hooks
    try:
        hooks.execute(sys.argv)
    except UnregisteredHookError as hook_error:
        if ansible_failed:
            raise hook_error
